library(dataresqc)
library(stats)
library(readxl)
library(gdata)
library(lubridate)
library(splitstackshape)

#####
# READ, FORMAT FILES, AND DESEASONALISE TEMPERATURE SERIES
#####

df <- data.frame(read_xlsx("~/paper2/ARM_stations.xlsx", skip=8, col_names=T))
# ignore validation stations and others that won't be used
df <- df[!grepl("CAD|HRL|MSK|TMS|ZUR",df[,3]),]

histpath <- "~/paper2/Data/Preprocessed/1_Homogenised/hist/"
refpath_ta <- list.files("~/paper2/Data/Preprocessed/3_Offset/ref/", full.names = T)
refpath_slp <- list.files("~/paper2/Data/Preprocessed/1_Homogenised/ref/",full.names = T)[grepl("_slp.",list.files("~/paper2/Data/Preprocessed/1_Homogenised/ref/"))]
refpath <- c(refpath_ta,refpath_slp)

# read hist series
histsef.files <- list.files(histpath, full.names=T)
sfnam <- paste0(sapply(histsef.files, function(x){read_meta(x)["id"]}),"_",sapply(histsef.files, function(x){read_meta(x)["var"]}))
sefhist <- data.frame(matrix(unlist(lapply(histsef.files, function(x){read_sef(x, all=T)}$Value)), nrow=120, byrow = F))
names(sefhist) <- gsub("CHIMES_|GCOS_|KNMI_|Palatine-Society_|SRM_|Royal-Society_","",sfnam)
for (i in 1:length(names(sefhist))) {
  names(sefhist)[i] <- paste0(df[grepl(sub("_.*","",names(sefhist))[i],df[,1]),3],"_",sub(".*_","",names(sefhist))[i])
  }
hist <- sefhist[,order(names(sefhist))]

# read ref series
refsef.files <- refpath
rfnam <- paste0(sapply(refpath, function(x){read_meta(x)["name"]}),"_",sapply(refpath, function(x){read_meta(x)["var"]}))
sefref <- data.frame(matrix(unlist(lapply(refsef.files, function(x){read_sef(x, all=T)}$Value)), nrow=25933, byrow = F))
names(sefref) <- rfnam
for (i in 1:length(names(sefref))) {
  names(sefref)[i] <- paste0(df[grepl(sub("_.*","",names(sefref))[i],df[,1]),3],"_",sub(".*_","",names(sefref))[i])
}
sefref <- sefref[,!(names(sefref) %in% c("GDK_ta","GDK_slp"))] # exclude data that haven't arrived yet
ref <- sefref[,order(names(sefref))]

print("filtering temperature seasonality")
sd <- as.Date("1950-01-01"); ed <- as.Date("2020-12-31")
date = seq(sd, ed, "days")
ref_ta <- ref[,grepl("_ta",colnames(ref))]

### make vector x with doy/ndoy
doy <- yday(seq(as.Date(date[1]),as.Date(date[length(date)]), by = "day"))
ndoy <- diff(seq(as.Date(date[1]), as.Date(paste((year(as.Date(date[length(date)]))+1),"-01-01",sep="")), by = "year"))
ndoy <- rep(ndoy,ndoy)
ndoy <- ndoy[1:length(doy)]

x <- doy/ndoy

# define seasonality function
rm.seas <- function(tempdata) {
  tempdata_cor <- tempdata
  
  temp.jg.cos <- cos(2*pi*x)
  temp.jg.sin <- sin(2*pi*x)
  temp.jg.cos2 <- cos(4*pi*x)
  temp.jg.sin2 <- sin(4*pi*x)
  
  temp.jg.lm <- lm(tempdata ~ temp.jg.cos + temp.jg.sin + temp.jg.cos2 + temp.jg.sin2)
  
  temp.jg.cor <- as.numeric(tempdata[as.numeric(names(temp.jg.lm$fitted.values))] - temp.jg.lm$fitted.values)
  tempdata_cor[!is.na(tempdata_cor)] <- temp.jg.cor
  return(tempdata_cor)  
}   

# calculate seasonality-corrected temperature data
ref_deseas <- apply(ref_ta,2,rm.seas)
# this procedure has removed the smoothed mean climatology for each temperature series

clim <- ref_ta - ref_deseas # to obtain mean seasonal climatology separately

ref[,colnames(ref)%in%colnames(ref_deseas)] <- ref_deseas
ref <- data.frame(date,ref)

# subtract seasonal component (with fitted values of reference data) from historical data; i.e. same sin/cos curve
hist_ta <- hist[,grepl("_ta",colnames(hist))]
hist_deseas <- hist_ta - clim[5784:5903,]
hist[,colnames(hist)%in%colnames(hist_deseas)] <- hist_deseas
hist <- cbind(date=seq(as.Date("1788-11-01"),as.Date("1789-02-28"),"days"),hist)
rm(doy,ndoy,x,rm.seas)


#####
# STANDARDISATION
#####

print("standardising data")
data_stand <- data.frame(scale(ref[,-1]))
ref_stand <- cbind(date,data_stand)
ref_stand[,-1] <- round(ref_stand[,-1],2)
rm(data_stand)
# file with deseasonalised and standardised data of reference period
write.table(ref_stand, file = "~/paper2/Data/Preprocessed/4_Deseas+Standardised/Ref_deseas_stand.txt", append = FALSE, sep = "\t", dec = ".",
            row.names = FALSE, col.names = TRUE)

# standardise short historical series with mean & standard deviation of 1950-2020 data
hist_stand <- (hist[,-1] - as.list(t(colMeans(ref[,-1],na.rm=TRUE)))) / apply(ref[,-1],2,sd,na.rm=TRUE)
ref <- ref[,-1]; hist <- hist[,-1]
for (i in 1:ncol(ref)) {
  new <- (hist[,i] - mean(ref[,i],na.rm=T)) / sd(ref[,i],na.rm=T)
  hist[,i] <- round(new,2)
}
hist_stand <- cbind(date=seq(as.Date("1788-11-01"),as.Date("1789-02-28"),"days"),hist)
# file with deseasonalised data of 1788/89 winter
write.table(hist_stand, file = "~/paper2/Data/Preprocessed/4_Deseas+Standardised/Hist_deseas_stand.txt", append = FALSE, sep = "\t", dec = ".",
            row.names = FALSE, col.names = TRUE)

#####
# ADD WTs
#####

wts1 <- read.table("~/paper2/Data/Preprocessed/WT_17632009.txt", header = T, sep = "")

wts2 <- read.table("~/paper2/Data/Preprocessed/WT_19572020.txt", header = T, sep = ";")[,-c(1:2)]
wts2 <- cbind(date=as.character(seq(as.Date('1957-09-01'),as.Date('2021-01-01'), by = "days")), wts2)
wts2 <- cbind(wts2, 1, data.frame(matrix(NA, 23134, 12)))
names(wts2) <- names(wts1)
wts2[,2] <- as.numeric(wts2[,2])

wts <- rbind(wts1, wts2[19116:23133,])
wts[,1] <- as.Date(wts[,1])
names(wts) <- c("date","WT_1","WT_1_prob","WT_2","WT_2_prob","WT_3","WT_3_prob","WT_4","WT_4_prob",
                "WT_5","WT_5_prob","WT_6","WT_6_prob","WT_7","WT_7_prob")

ref <- read.table("~/paper2/Data/Preprocessed/4_Deseas+Standardised/Ref_deseas_stand.txt", header = T, sep = "")
predictor <- cbind(date=ref[,1], wts[which(wts$date >= "1950-01-01" & wts$date <= "2020-12-31"), -1], ref[,-1])
write.table(predictor, file = "~/paper2/Data/Preprocessed/Predictor.txt", append = FALSE, sep = "\t", dec = ".",
            row.names = FALSE, col.names = TRUE)

hist <- read.table("~/paper2/Data/Preprocessed/4_Deseas+Standardised/Hist_deseas_stand.txt", header = T, sep = "")
predictand <- cbind(date=hist[,1], wts[which(wts$date >= "1788-11-01" & wts$date <= "1789-02-28"), -1], hist[,-1])
write.table(predictand, file = "~/paper2/Data/Preprocessed/Predictand.txt", append = FALSE, sep = "\t", dec = ".",
            row.names = FALSE, col.names = TRUE)
