library(ncdf4)
library(zoo)
library(tidyverse)

###################
### READ/FORMAT ZONAL MEANS AND FIT TREND LINES
###################
# set path and filename
ncpath <- "~/paper2/Data/Preprocessed/era5_t2m_zonmean_daymean_19502020_masked.nc"
dname <- "t2m"

# open a netCDF file
ncin <- nc_open(ncpath)
print(ncin)

# get longitude and latitude
lon <- ncvar_get(ncin,"lon")
nlon <- dim(lon)
head(lon)
lat <- ncvar_get(ncin,"lat")
nlat <- dim(lat)
head(lat)

# get temperature
tmp_array <- ncvar_get(ncin,dname)
dlname <- ncatt_get(ncin,dname,"long_name")
dunits <- ncatt_get(ncin,dname,"units")
fillvalue <- ncatt_get(ncin,dname,"_FillValue")
dim(tmp_array)

# replace netCDF fill values with NA's
tmp_array[tmp_array==fillvalue$value] <- NA
length(na.omit(as.vector(tmp_array)))

# colnames
tmp_array <- round(t(tmp_array), 4)
colnames(tmp_array) <- round(lat, 4)
tmp_array <- tmp_array - 273.15     # Kelvin to Celsius

# dates
sd = as.Date("1950-01-01"); ed = as.Date("2020-12-31")
dates <- seq(sd, ed, "days")

# add dates to data.frame, which is reduced to latitudes of interest
zonmean_ta <- cbind.data.frame(dates, tmp_array)

# regression to fit trend to each latitude
timesteps <- 1:nrow(zonmean_ta)
fit <- list()
for (i in 2:ncol(zonmean_ta)) {
  fit[[i]] <- lm(zonmean_ta[,i] ~ timesteps)
}
fit <- fit[-1]

# extend prediction until 2020 and center fitted values around mid-period 1985-01
fit_extend <- list(); fit_vals85 <- list()
for (i in 1:length(fit)) {
  fit_extend[[i]] <- predict(fit[[i]], data.frame(timesteps))
  fit_vals85[[i]] <- fit_extend[[i]] - fit_extend[[i]][grepl("1985-01-01", zonmean_ta$dates)]
}
fit_vals85 <- data.frame(dates, Reduce(cbind, fit_vals85))
names(fit_vals85) <- colnames(zonmean_ta)

df <- fit_vals85[,c(1,42:158)]

###################
### WRITE NEW NETCDF FILE OF FITTED VALUES TO BE LATER SUBTRACTED FROM E-OBS DATASET WITH CDO
###################
filename="slope_trends_lats.nc"

xvals <- seq(0, 0, 0)
yvals <- seq(36, 65, 0.25) 
nx <- length(xvals)
ny <- length(yvals)
lon1 <- ncdim_def("longitude", "degrees_east", xvals)
lat2 <- ncdim_def("latitude", "degrees_north", yvals)

t <- ncdim_def( "Time", "days since 1949-12-31", 1:length(timesteps), unlim=TRUE)
timevar <- ncvar_def(name = "time",
                     units = 'days since 1950-01-01 00:00:00',
                     dim = list(t))
variables <- list(timevar)
mv <- -999 #missing value to use
var_temp <- ncvar_def("temperature", "celsius", list(lon1,lat2,t), 
            longname="Fitted Values of Slopes for latitudes 36-65N over 1950-2020, calculated from ERA5", mv) 

ncnew <- nc_create(filename, list(var_temp))
df2 <- t(df[,-1]); df_nc <- df2[order(nrow(df2):1),]
ncvar_put(ncnew, var_temp, df_nc, count=c(1,117,25933))



# Don't forget to close the file
nc_close(ncnew)



### TRY EXTEND THIS IN 3RD DIMENSION (ALONG LONGITUDE)
# filename="slope_trends_around84_ekf400_3D.nc"
# 
# xvals <- seq(-9.95013956920733, 39.9498602311671, 0.0999999995999487)
# yvals <- seq(37.0498605720475, 64.9498604835084, 0.0999999996826557) 
# nx <- length(xvals)
# ny <- length(yvals)
# lon1 <- ncdim_def("longitude", "degrees_east", xvals)
# lat2 <- ncdim_def("latitude", "degrees_north", yvals)
# 
# days <- c(1:25202)
# t <- ncdim_def( "Time", "days since 1949-12-31", 1:length(days), unlim=TRUE)
# timevar <- ncvar_def(name = "time",
#                      units = 'days since 1950-01-01 00:00:00',
#                      dim = list(t))
# variables <- list(timevar)
# mv <- -999 #missing value to use
# var_temp <- ncvar_def("temperature", "celsius", list(lon1,lat2,t), longname="Fitted Values of Slopes for latitudes 37N-65N over 1950-2018, taken from EKF400v2", mv) 
# 
# ncnew <- nc_create(filename, list(var_temp))
# ncvar_put(ncnew, var_temp, replicate(500, t(daily_fit[,2:281])), count=c(500,280,25202))
# 
# # Don't forget to close the file
# nc_close(ncnew)
# 
