library(dataresqc)
library(ncdf4)
library(zoo)

###################
### READ/FORMAT ZONAL MEANS AND FIT TREND LINES
###################
# set path and filename
ncpath <- "~/paper2/Data/Preprocessed/era5_t2m_zonmean_daymean_19502020_masked.nc"
dname <- "t2m"

# open a netCDF file
ncin <- nc_open(ncpath)
print(ncin)

# get longitude and latitude
lon <- ncvar_get(ncin,"lon")
nlon <- dim(lon)
head(lon)
lat <- ncvar_get(ncin,"lat")
nlat <- dim(lat)
head(lat)

# get temperature
tmp_array <- ncvar_get(ncin,dname)
dlname <- ncatt_get(ncin,dname,"long_name")
dunits <- ncatt_get(ncin,dname,"units")
fillvalue <- ncatt_get(ncin,dname,"_FillValue")
dim(tmp_array)

# replace netCDF fill values with NA's
tmp_array[tmp_array==fillvalue$value] <- NA
length(na.omit(as.vector(tmp_array)))

# colnames
tmp_array <- round(t(tmp_array), 4)
colnames(tmp_array) <- round(lat, 4)
tmp_array <- tmp_array - 273.15     # Kelvin to Celsius

# dates
sd = as.Date("1950-01-01"); ed = as.Date("2020-12-31")
dates <- seq(sd, ed, "days")

# add dates to data.frame, which is reduced to latitudes of interest
zonmean_ta <- cbind.data.frame(dates, tmp_array)

# regression to fit trend to each latitude
timesteps <- 1:nrow(zonmean_ta)
fit <- list()
for (i in 2:ncol(zonmean_ta)) {
  fit[[i]] <- lm(zonmean_ta[,i] ~ timesteps)
}
fit <- fit[-1]
names(fit) <- colnames(zonmean_ta)[-1]


###################
### DETREND 'STATION' DATA AROUND MID-PERIOD (1985)
###################

refpath <- "~/paper2/Data/Preprocessed/1_Homogenised/ref/"
ref.files <- list.files(refpath, full.names = T)[grepl("_ta.tsv|_ta_hom.tsv", list.files(refpath))]

for (k in 1:length(ref.files)) {
  ref <- read_sef(ref.files[k], all=T)
  meta <- data.frame(t(read_meta(ref.files[k])))
  
  # find closest ekf latitude to station latitude
  ind <- as.numeric(which.min(abs(as.numeric(meta$lat) - as.numeric(names(fit)))))
  
  # extend prediction until 2020
  fit_vals <- predict(fit[[ind]], data.frame(timesteps))
  
  # fitted values centred around mid-period 1985-01
  fit_vals85 <- fit_vals - fit_vals[grepl("1985-01-01", zonmean_ta$dates)]
  
  # ERA5 fitted values of zonal mean lm applied to ref
  daily_fit_vals = data.frame(dates, fit_vals85)                            
  ref$detrended <- round(ref$Value - daily_fit_vals$fit_vals85, 2)
  
  newref.path <- "~/paper2/Data/Preprocessed/2_Detrended/ref/"
  
  for (n in 1:nrow(ref)) {
    if (ref$Meta[n] == "") {
      ref$Meta[n] <- paste0("detrend=",round((-1)*daily_fit_vals$fit_vals85[n], 2))
    } else {
      ref$Meta[n] <- paste0(ref$Meta[n],"|detrend=",round((-1)*daily_fit_vals$fit_vals85[n], 2))
    }
  }
  
  # write new SEF files
  write_sef(Data = ref[,c(2:6,10)],
            outpath = newref.path,
            variable = meta["var"],
            cod = meta["id"],
            nam = meta["name"],
            lat = meta["lat"],
            lon = meta["lon"],
            alt = meta["alt"],
            link = meta["link"],
            sou = meta["source"],
            units = meta["units"],
            stat = meta["stat"],
            metaHead = meta["meta"],
            meta = ref$Meta,
            period = "day",
            time_offset = 0,
            keep_na = TRUE,
            note="detrend")
}

#plot(ref$Value,type="l")
#plot(ref$detrended,type="l")
