# This script's goal is to homogenise %ta and %slp historical data (1788-1789) by assessing the refence period data's (1950-2020)
# relation to EKF400 monthly values. The assumption being that the EKF400 is well-reconstructed and preserves the
# long-term picture/development/variability. To compare and calculate offsets from the EKF400 curves, monthly means are
# computed from the daily 'observations' (hist & ref). After the historical monthly data is 'adjusted', I re-add the daily variability.

# The steps in the script are the following:
#
# 1. calculate monthly means from daily data and
#    format hist, ref, and ekf400 data so they are comparable
# 2. calculate offsets for each station and variable based on
#    mean difference ekf400-ref for NOV,DEC,JAN,FEB
# 3. adjust hist so that they have same offsets as ref
# 4. add back daily variability to adjusted hist


library(dataresqc)
library(readxl)
library(lubridate)
library(zoo)

count_days <- function(x) {
  n <- length(x)
  if (n >= 18) {
    x <- c(0, x)
    fdiff <- x[2:(n+1)]-x[1:n]
    if (max(fdiff) > 5) n <- 0
  }
  return(n)
}

#####
# CALCULATE MONTHLY MEANS FOR HISTORICAL AND REFERENCE DATA (OBSERVATIONS)
#####

df <- data.frame(read_xlsx("~/paper2/ARM_stations.xlsx", skip=8, col_names=T))
# ignore validation stations and others that won't be used
df <- df[!grepl("HRL|MSK|TMS|ZUR",df[,3]),]

historig <- "~/paper2/Data/Unprocessed/Historical/3_SEF-daily/"
reforighom <- "~/paper2/Data/Preprocessed/1_Homogenised/ref/"

# read hist series
histsef.files <- list.files(historig, full.names=T)
sfnam <- paste0(sapply(histsef.files, function(x){read_meta(x)["id"]}),"_",sapply(histsef.files, function(x){read_meta(x)["var"]}))
sefhist <- lapply(histsef.files, function(x){read_sef(x, all=T)})
metahist <- lapply(histsef.files, function(x){read_meta(x)})
names(sefhist) <- gsub("CHIMES_|GCOS_|KNMI_|Palatine-Society_|SRM_|Royal-Society_","",sfnam)
names(metahist) <- gsub("CHIMES_|GCOS_|KNMI_|Palatine-Society_|SRM_|Royal-Society_","",sfnam)
for (i in 1:length(names(sefhist))) {
  names(sefhist)[i] <- paste0(df[grepl(sub("_.*","",names(sefhist))[i],df[,1]),3],"_",sub(".*_","",names(sefhist))[i])
  names(metahist)[i] <- paste0(df[grepl(sub("_.*","",names(metahist))[i],df[,1]),3],"_",sub(".*_","",names(metahist))[i])
}

# read ref series
refsef.files <- list.files(reforighom, full.names=T)
rfnam <- paste0(sapply(refsef.files, function(x){read_meta(x)["name"]}),"_",sapply(refsef.files, function(x){read_meta(x)["var"]}))
sefref <- lapply(refsef.files, function(x){read_sef(x, all=T)})
names(sefref) <- rfnam
for (i in 1:length(names(sefref))) {
  names(sefref)[i] <- paste0(df[grepl(sub("_.*","",names(sefref))[i],df[,1]),3],"_",sub(".*_","",names(sefref))[i])
}

# calc means
momeanhist <- matrix(unlist(lapply(sefhist, function(m){
  xmonthly <- aggregate(m$Value, list(m$Year,m$Month), mean)
  xmonthly <- xmonthly[order(xmonthly[,1],xmonthly[,2]),]
  nmonthly <- aggregate(m$Day, list(m$Year,m$Month), count_days)
  nmonthly <- nmonthly[order(nmonthly[,1],nmonthly[,2]),]
  nmonthly$miss <- days_in_month(as.Date(paste(nmonthly[,1],nmonthly[,2],1,sep="-"))) - nmonthly$x
  xmonthly$x[which(nmonthly$miss>=11)] <- NA
  return(xmonthly$x)
  })), nrow=4, byrow = F)
colnames(momeanhist) <- names(sefhist)

momeanref <- matrix(unlist(lapply(sefref, function(m){
  xmonthly <- aggregate(m$Value, list(m$Year,m$Month), mean)
  xmonthly <- xmonthly[order(xmonthly[,1],xmonthly[,2]),]
  nmonthly <- aggregate(m$Day, list(m$Year,m$Month), count_days)
  nmonthly <- nmonthly[order(nmonthly[,1],nmonthly[,2]),]
  nmonthly$miss <- days_in_month(as.Date(paste(nmonthly[,1],nmonthly[,2],1,sep="-"))) - nmonthly$x
  xmonthly$x[which(nmonthly$miss>=11)] <- NA
  return(xmonthly$x)
  })), nrow=852, byrow = F)
colnames(momeanref) <- names(sefref)

momeanhist <- data.frame(momeanhist); momeanhist <- momeanhist[,order(names(momeanhist))]
momeanref <- data.frame(momeanref); momeanref <- momeanref[,order(names(momeanref))]
momeanref <- momeanref[,!grepl("GDK_ta|GDK_slp|HMB_ta",colnames(momeanref))] #exclude data not yet received
momeanref <- momeanref[-which(2 < rep(1:12,71) & 11 > rep(1:12,71)),] # select winter months of interest NDJF
momeanref <- momeanref[1:216,] # select years relevant to ekfref 1950-2003
    
#####
# READ EKF400 MONTHLIES AND FORMAT SAME AS ABOVE
#####

histekf <- "~/paper2/ekf/hist/"
refekf <- "~/paper2/ekf/ref/"

# contain only winter months of interest NDJF
histekf.files <- list.files(histekf, full.names=T)
ekfnam <- gsub(".{8}$", "", gsub(".*/","",histekf.files))
ekfhist <- lapply(histekf.files, function(x){read.table(x, header=F)[-c(1:2,7:8),4]})
names(ekfhist) <- ekfnam
ekfhist <- ekfhist[names(ekfhist)[which(names(ekfhist) %in% names(sefhist))]]
momeanhist_ekf <- matrix(unlist(ekfhist), nrow=4, byrow = F)
colnames(momeanhist_ekf) <- names(ekfhist)

refekf.files <- list.files(refekf, full.names=T)
ekfnam <- gsub(".{8}$", "", gsub(".*/","",refekf.files))
ekfref <- lapply(refekf.files, function(x){read.table(x, header=F)[,4]})
names(ekfref) <- ekfnam
ekfref <- ekfref[names(ekfref)[which(names(ekfref) %in% names(sefref))]]
momeanref_ekf <- matrix(unlist(ekfref), nrow=216, byrow = F)
colnames(momeanref_ekf) <- names(ekfref)

momeanhist_ekf <- data.frame(momeanhist_ekf)
momeanref_ekf <- data.frame(momeanref_ekf); momeanref_ekf <- momeanref_ekf[,!grepl("GDK_ta|GDK_slp",colnames(momeanref_ekf))]

# Pa to hPa
momeanhist_ekf[,c(names(momeanhist_ekf)[grepl("_slp",names(momeanhist_ekf))])] <-
  momeanhist_ekf[,c(names(momeanhist_ekf)[grepl("_slp",names(momeanhist_ekf))])] / 100
momeanref_ekf[,c(names(momeanref_ekf)[grepl("_slp",names(momeanref_ekf))])] <-
  momeanref_ekf[,c(names(momeanref_ekf)[grepl("_slp",names(momeanref_ekf))])] / 100


#####
# CALCULATE DIFFERENCES AND AVERAGE FRO NOV THROUGH FEB (BOTH ref AND hist)
#####

h.diff <- data.frame(month=c(11,12,1,2), momeanhist_ekf - momeanhist)

r.diff <- data.frame(month=rep(c(1,2,11,12),54), momeanref_ekf - momeanref)
Nov.r.diff = t(data.frame(colMeans(r.diff[which(r.diff$month==11),-1],na.rm=T)))
Dec.r.diff = t(data.frame(colMeans(r.diff[which(r.diff$month==12),-1],na.rm=T)))
Jan.r.diff = t(data.frame(colMeans(r.diff[which(r.diff$month==1),-1],na.rm=T)))
Feb.r.diff = t(data.frame(colMeans(r.diff[which(r.diff$month==2),-1],na.rm=T)))
r.diff <- as.data.frame(rbind(Nov.r.diff,Dec.r.diff,Jan.r.diff,Feb.r.diff))
rownames(r.diff) <- c(11,12,1,2)


adjust <- cbind(h.diff$month, r.diff - h.diff[,-1])
# amount that I have to add to 'hist' so that it has same monthly distance of 'ref' to 'ekf400'

test <- momeanhist - adjust[,-1]
verify <- momeanhist_ekf - test
# compare 'test' and 'r.diff.means' to see if they have same values, i.e. same offset to ekf400
round(verify - r.diff, 1) # should all be = 0


#####
# ADJUST HIST SERIES AND APPLY 'OFFSET' TO DAILIES
#####

adjust.daily <- data.frame(rbind(adjust[rep(1, each=30),],
                                    adjust[rep(2, each=31),],
                                      adjust[rep(3, each=31),],
                                        adjust[rep(4, each=28),]))[,-1]

xnew.path <- "~/paper2/Data/Preprocessed/1_Homogenised/hist/"
sefhist <- sefhist[order(names(sefhist))]
metahist <- metahist[order(names(metahist))]

for (i in 1:length(sefhist)) {
  x <- sefhist[[i]][,c(2:6,8:9)]
  meta <- metahist[[i]]
  xnew <- x
  xnew$Value <- x$Value - round(adjust.daily[,i],2)
  
  write_sef(Data = xnew,
            outpath = xnew.path,
            variable = meta["var"],
            cod = meta["id"],
            nam = meta["name"],
            lat = meta["lat"],
            lon = meta["lon"],
            alt = meta["alt"],
            link = meta["link"],
            sou = meta["source"],
            units = meta["units"],
            stat = meta["stat"],
            metaHead = meta["meta"],
            meta = paste0(xnew$Meta,"|homREFadj=",round(adjust.daily[,i],2)),
            period = "day",
            time_offset = 0,
            keep_na = TRUE,
            note="homref")
  
  lines(xnew$Value,type="l")
}


# xmonthly <- aggregate(m$Value, list(m$Year,m$Month), mean)
# xmonthly <- xmonthly[order(xmonthly[,1],xmonthly[,2]),]
# 
# round(momeanhist_ekf$ZWA_ta - xmonthly$x - r.diff$ZWA_ta, 1)
# r.diff$ZWA_ta
# plot(sefhist$PET_ta$Value-adjust.daily$PET_ta+sefhist$MOS_ta$Value+adjust.daily$MOS_ta,type="l")
