# "03" Analogs
# analogue resampling method (ARM)
# Duncan's version, July 2021

### input needed:
# I performed steps that are in the pre-processing script separately, what is needed here are the predictor and predictand tables/files
# from masterscript:  specifications
### options: 
# select distance measure, number of analogues, seasonal window, 
# time window around target for leave-one-out validation
### packages: 
# lubridate, doParallel, purrr, hydroGOF, stats
# parallelisation: default is set to use 80% of all available cores

#############################################################################################################
#############################################################################################################
# 1. DEFINE PREDICTOR AND PREDICTAND TABLES
# 2. DEFINE DISTANCE FUNCTION
# 3. ANALOGUE CALCULATION
#    2.1 DEFINE TARGET AND ANALOGUE POOL
#    2.2 DEFINE ANALOGUE FUNCTIONS
#    2.3 CALCULATE ANALOGUE DATES AND DISTANCES
# 4. WRITE TABLES
#############################################################################################################
#############################################################################################################


#############################################################################################################
### 1. DEFINE PREDICTOR AND PREDICTAND TABLES
#############################################################################################################

ref <- read.table("~/paper2/Data/Preprocessed/Predictor.txt", header=TRUE, sep="")
ref$date <- as.Date(ref$date, "%Y-%m-%d")
hist <- read.table("~/paper2/Data/Preprocessed/Predictand.txt", header=TRUE, sep="")
hist$date <- as.Date(hist$date, "%Y-%m-%d")
data <- rbind(hist,ref)

### select weather types according to masterscript
wt_cols <- colnames(data)[grepl("WT",colnames(data)) & !grepl("prob",colnames(data))]
wt_prob_cols <- colnames(data)[grepl("WT",colnames(data)) & grepl("prob",colnames(data))]

if(is.numeric(specifications$WT_sel)&specifications$WT_sel>=1){
  data <- data[,!colnames(data)%in%wt_cols[!wt_cols%in%wt_cols[c(1:specifications$WT_sel)]] &
                !colnames(data)%in%wt_prob_cols]
} else {
  if(is.numeric(specifications$WT_sel)&specifications$WT_sel<1){
    wt_prob <- data[,wt_prob_cols]
    wt_prob_sum <- data.frame(matrix(NA,ncol=length(wt_prob_cols),nrow = nrow(data)))
    wt_prob_sum <- sapply(1:length(wt_prob_cols),function(i){wt_prob_sum[,i]<-rowSums(wt_prob[1:i])})
    
    for(i in 1:nrow(data)){
      data[i,wt_cols][-c(1:min(which(wt_prob_sum[i,]>specifications$WT_sel)))]<-NA
    }
    data<-data[,!colnames(data)%in%wt_prob_cols]
  }}

rm(wt_prob, wt_prob_sum)


# to reflect different station networks for 1788 and 1789
data[grepl("-11-|-12-",data$date),grepl("KEZ",names(data))] <- NA
data[grepl("-01-|-02-",data$date),grepl("BRL|COP|ERF|LNB|MID|STR|SPT",names(data))] <- NA

predictand <- data[grepl("-11-|-12-|-01-|-02-", data$date),]
predictor <- data[!grepl("1788|1789|-04-|-05-|-06-|-07-|-08-|-09-", data$date),]
predictor <- predictor[,!colnames(predictor)%in%wt_cols[!wt_cols%in%wt_cols[1]]]
#data <- rbind(predictand,predictor)[,-c(7,8)]

# remove variables
rm(wt_cols,wt_prob_cols,i)


# get time
ptm <- proc.time()

#############################################################################################################
### 2. DEFINE DISTANCE FUNCTIONS
#############################################################################################################
print(paste("define distance function:",specifications$distance))

if(specifications$distance=="Euclid") {
  # distance function: euclidean distance
  # NA's are removed, if all NA values -> NA
  # distance is weighted with inverse number of parameters (penalty for less parameters available)
  # similar to RMSE which weights by square root of inverse number of parameters (weaker penalty)
  distance <- function(predictor, target) {
    if (all(is.na(target))){return(NA)}
    else {sqrt(sum((target - predictor) ^ 2, na.rm = T))/length(which(!is.na(predictor)))}
  }
} else {
if(specifications$distance=="RMSE") {
  distance <- function(predictor, target) {
    if (all(is.na(target))){return(NA)}
    else {rmse(target,predictor)}
  }
} else {
### Mahalanobis distance: not good, should be separated for individual groups of variables or stations
 if(specifications$distance=="Mahalanobis") {
  cov.mat <- cov(data[,-c(1,2)], use = "pairwise.complete.obs")
  distance <- function(predictor, target) {
    if (all(is.na(target))){return(NA)}
    else {
    sqrt(mahalanobis(target,predictor,cov=cov.mat))/length(which(!is.na(predictor)))}
  }
} else {
### MAE
 if(specifications$distance=="MAE") {
  distance <- function(predictor, target) {
    if (all(is.na(target))){return(NA)}
    else {
    mean(abs(target-predictor))/length(which(!is.na(predictor)))}
  }
}}}}


#############################################################################################################
### 3. ANALOGUE CALCULATION
#############################################################################################################

#------------------------------------------------------------------------------------------------------------
### 3.1 DEFINE TARGET AND ANALOGUE POOL
#------------------------------------------------------------------------------------------------------------
Analogpool <- predictor
rownames(Analogpool)<-Analogpool$date
Target <- predictand
rownames(Target)<-Target$date
Target <- split(Target,Target$date)

#------------------------------------------------------------------------------------------------------------
### 3.2 DEFINE ANALOGUE FUNCTIONS
#------------------------------------------------------------------------------------------------------------
print("define analogue function")

## outcome is a list of days with "nAnalog" best analogue dates and corresponding distances
Analogs = function(target,nAnalog,dayselection,window) {
  # define date to reconstruct and vector of weather types
  day <- target[["date"]]
  w_type <- as.numeric(target[grepl("WT",names(target)) & !grepl("prob",names(target))])
  
  # create matrix with measurements of target day
  target_data <- target[names(target)[!grepl("WT",names(target)) & !grepl("date",names(target))]]

  # make preselection to select only days +-XXX days from target date
  preselection <- sapply(seq(specifications$startRef, specifications$endRef), 
                         function(i) seq(as.Date(yday(day)-dayselection, origin=paste(i,"01-01",sep="-")),
                                         as.Date(yday(day)+dayselection, origin=paste(i,"01-01",sep="-")),
                                         by="day"))
  preselection <- c(as.Date(preselection,origin="1970-01-01"))

  # # if predictand within predictor period: remove predictand dates +-window from predictor
  pres_window <- seq(as.Date(day)-window, as.Date(day)+window, by="day")
  preselection <- preselection[!preselection%in%pres_window]
   
  preselection <- preselection[preselection >= as.Date(paste(specifications$startRef, 1, 1, sep="-"))]
  preselection <- preselection[preselection <= as.Date(paste(specifications$endRef, 12, 31, sep="-"))]
  
  # select possible analogues according to weather types and "preselection"
  analog_selection <- Analogpool[Analogpool$WT%in%w_type & Analogpool$date%in%preselection,]

  # calculate distances and get date/distance matrices
  dist<-apply(analog_selection[,-c(1:2)],1,distance,target=as.numeric(target_data))
  dist.min <- sort(dist, decreasing = F)
  dist.min <- dist.min[1:nAnalog]
  
  dist.index <- as.Date(names(dist.min))
  
  dist <- as.numeric(dist.min)
  
  # create list with analog dates and distances for target day
  analogue.ls <- list("analog.dates"=as.numeric(dist.index),"analog.dist"=dist)
  
  return(analogue.ls)
}

#------------------------------------------------------------------------------------------------------------
### 3.3 CALCULATE ANALOGUE DATES AND DISTANCES
#------------------------------------------------------------------------------------------------------------
print(paste("calculate best",specifications$nAnalog,"analogues"))

### create list of days, containing  analogue dates and distances
analogue.list<- mcmapply(Analogs,Target,nAnalog=specifications$nAnalog,
                    dayselection=specifications$dayselection,window=specifications$window, SIMPLIFY = F,
                    mc.cores = round(0.1*detectCores(),0))

### extract analogue dates from list as matrix
analog.dates<-data.frame(as.Date(names(analogue.list)),t(modify_depth(analogue.list,1,~ .[["analog.dates"]]) %>% as.data.frame))
colnames(analog.dates)<-c("date",1:specifications$nAnalog)
rownames(analog.dates)<-c(1:nrow(analog.dates))

### extract analogue distances from list as matrix
analog.dist<-data.frame(as.Date(names(analogue.list)),t(modify_depth(analogue.list,1,~ .[["analog.dist"]]) %>% as.data.frame))
colnames(analog.dist)<-c("date",1:specifications$nAnalog)
rownames(analog.dist)<-c(1:nrow(analog.dist))
     
                     
#############################################################################################################
### 4. WRITE TABLES
#############################################################################################################
print("write files to folder")

## create output folder for validation
subDir <- paste("ARM_run",Sys.Date(),sep="_")
dir.create(file.path("~/paper2/output",subDir), showWarnings = F)
outfolder <- paste(file.path("~/paper2/output",subDir),"/",sep="")
rm(subDir)

## save tables (dates, distances, readme)
if (specifications$seasonality==T) {props1 <- "deseas_"} else {props1 <- c()}
if (specifications$trend!=F) {props2 <- "detrend_"} else {props2 <- c()}
props3 <- specifications$distance
sdate <- as.character(analog.dates[1,1])
edate <- as.character(analog.dates[nrow(analog.dates),1])

write.table(analog.dates, file=paste0(outfolder,Sys.Date(),"_analog.dates_",sdate,"-",edate,"_",props1,props2,props3,".txt",sep=""))
write.table(analog.dist,file=paste0(outfolder,Sys.Date(),"_analog.dist_",sdate,"-",edate,"_",props1,props2,props3,".txt",sep=""))
readme<-data.frame("description"=c("analog calculation",paste(names(specifications),":",sep="")),"parameters"=NA)
readme[1,2]<-as.character(as.Date(Sys.Date()))
for(i in 2:nrow(readme)) {readme[i,2]<-as.character((paste(specifications[[i-1]], collapse = ", ")))}
readme[2,2]<- paste(unique(substr(colnames(predictand),1,3))[-c(1,2)],collapse = ", ")
write.table(readme,file=paste0(outfolder,Sys.Date(),"_readme_analogs_",sdate,"-",edate,"_",props1,props2,props3,".txt",sep=" "))


#############################################################################################################
## return summary
print(paste("analogue reconstruction from",sdate,"to",edate,sep=" "))
etm <- proc.time()
print(paste("calculation time:",(etm[3]-ptm[3])%/%60,"minutes", round((etm[3]-ptm[3])%%60),"seconds" ,sep=" "))
print(paste("files are in:",outfolder,sep=" "))

## remove variables
rm(Analogpool, Target, props1, props2, props3, data, sdate, edate, ptm, etm, i, 
   outfolder, Analogs, distance, readme, analogue.list)
if(specifications$distance=="Mahalanobis"){rm(cov.mat)}

# END
