### 05 EnKF
### Ensemble Kalman fitting (SLP)
### 17-08-2021

### input needed:
# analogue dates (with n best analogues); should be an ensemble of min 30 Analogues
# 3 pressure raster datasets (era5, era5_anom, era5_climatology)
# station measurements (TOT)
# station coordinates ("Data/Station_data/coordinates.txt")
# monthly difference between station measurements and corresponding gridcell of ERA5
# altitude data for vertical localisation
### options:
# vertical and horizontal localisation (+ cutoff distances), station selection, set observation error R
### packages:
# raster, gdata

#############################################################################################################
# from console (per year): Rscript 05_EnKF.R 1788
# for many years, use bash script for parallel runs (./EnKF_parallel_SLP.sh)
#############################################################################################################

#############################################################################################################
#############################################################################################################
# 1. READ DATA
#    1.1 READ ANALOGUE DATES
#    1.2 READ RASTER DATA
#    1.3 READ STATION DATA
# 2. KALMAN FITTING
#    2.1 DEFINE VALUES
#    2.2 CALCULATE DISTANCE MATRIX
#    2.3 Kalman Filter for each day in analogue dates
# 3. WRITE RESULTS TO FILES
#############################################################################################################
#############################################################################################################

# get time
ptm <- proc.time()
#in_year=1788
### the following lines take arguments (year) from console
args <- commandArgs(TRUE)
if (length(args)>0) {
  in_year = as.numeric(args[1])
  
  # load other scripts
  source("D01_packages.R")
  
  # define kalmanset
  kalmanset <- list(
    analog_dates = "output/ARM_run_2021-09-06/2021-09-06_analog.dates_1788-11-01-2020-12-31_deseas_detrend_RMSE.txt", # define analog dates
    station_data = "input/TOT_homhist.txt",      # define station data (TOT)
    station_xy = "input/coordinates.xlsx", # define station coordinate file
    input_raster = "grid/era5_slp.nc", # define input raster (era5, rest calculated with cdo commands)
    input_raster_anom = "grid/era5_anom.nc",
    climatology = "grid/era5_timmean.nc",
    # define set of stations to do kalman filter
    #stationset = c("Nov-Dec"), # define set of stations to do kalman filter
    stationset = c("Jan-Feb"),
    localize = T,
    cutoff = 1500, # define horizontal cutoff distance L [km] (if localize = T)
    Z_cutoff = 3000, # define vertical cutoff distance z [m]  (if localize = T)
    R = 3 # define measurement error R [hPa]
  )
}

if (!exists("in_year")) message("define year for which EnKF should be run")
stopifnot(exists("in_year"))

print(in_year)

## create output folder for validation
subDir <- paste("EnKF",Sys.Date(),sep="_slp_")
dir.create(file.path(dirname(kalmanset$analog_dates),subDir), showWarnings = F)
outfolder <- paste(file.path(dirname(kalmanset$analog_dates),subDir),"/",sep="")
rm(subDir)

#############################################################################################################
### 1. READ DATA
#############################################################################################################

#----------------------------------------------------------------------------------------
### 1.1 READ ANALOGUE DATES
#----------------------------------------------------------------------------------------
### load analogue dates
analog.dates <- read.table(kalmanset$analog_dates,as.is=T)

#----------------------------------------------------------------------------------------
### 1.2 READ RASTER DATA
#----------------------------------------------------------------------------------------
### load input rasters
suppressWarnings(tabs <- brick(kalmanset$input_raster))
suppressWarnings(tabs_anom <- brick(kalmanset$input_raster_anom))
suppressWarnings(tabs_climatology <- brick(kalmanset$climatology))

climatology_vals <- rasterToPoints(tabs_climatology)
substr(dimnames(climatology_vals)[[2]][3:368],2,5) <- "0000"

EU.alt <- raster("grid/alt_era5.nc")

#----------------------------------------------------------------------------------------
### 1.3 READ STATION DATA
#----------------------------------------------------------------------------------------
if (kalmanset$stationset=="Nov-Dec") { 
    stationset <- c("BAR","BAS","BRL","BER","BOL","BSW","BUD","CBT","CET","CLF","COP","EDI","ERF","GEN","GOT","GDC",
                    "HOH","INN","ROC","LIL","LIV","LDN","LNB","MAD","MAN","MAR","MID","MIL","MNT","MOS","NAN","PAD",
                    "PAR","PRK","REG","RXV","ROM","RST","ROU","STB","SPT","PET","STO","STR","TRH","TRK","UPS","VIE","VIL","WAR",
                    "ZAG","ZWA")
  } else if (kalmanset$stationset=="Jan-Feb") {
    stationset <- c("BAR","BAS","BER","BOL","BSW","BUD","CBT","CET","CLF","EDI","GEN","GOT","GDC",
                    "HOH","INN","KEZ","ROC","LIL","LIV","LDN","MAD","MAN","MAR","MIL","MNT","MOS","NAN","PAD",
                    "PAR","PRK","REG","RXV","ROM","RST","STB","PET","STO","TRH","TRK","UPS","VIE","VIL","WAR",
                    "ZAG","ZWA")  
  } else {
  stationset <- kalmanset$stationset
}


### Read station pressure measurements
TOT <- read.table(kalmanset$station_data,header=T,as.is=T)
station_data <- data.frame("date"=as.Date(TOT$date),TOT[,names(TOT)%in%as.vector(matchcols(TOT, with="slp", method = "or"))])
station_data <- data.frame("date"=station_data$date,station_data[,names(station_data)%in%as.vector(matchcols(station_data, with=stationset, method = "or"))])
stationset <- substr(colnames(station_data[,-1]),1,3)

#------------------------------------------------------------------------------------------------
## for console run
if (exists("in_year")) {
  if (kalmanset$stationset=="Jan-Feb") {
    station_data <- station_data[which(year(station_data$date)%in%in_year & grepl("-01-|-02-",station_data$date)),]
    analog.dates <- analog.dates[which(substr(analog.dates$date,1,4)%in%in_year & grepl("-01-|-02-",analog.dates$date)),]
  } else if (kalmanset$stationset=="Nov-Dec") {
    station_data <- station_data[which(year(station_data$date)%in%in_year & grepl("-11-|-12-",station_data$date)),]
    analog.dates <- analog.dates[which(substr(analog.dates$date,1,4)%in%in_year & grepl("-11-|-12-",analog.dates$date)),]
  }
}
#------------------------------------------------------------------------------------------------

### Read station coordinates
station_coords <- read_excel(kalmanset$station_xy,col_names=TRUE,skip=0) # make sure readxl package is loaded
station_coords <- as.data.frame(station_coords[which(station_coords$abbr.%in%stationset),])

### remove bias between pressure measurements and gridcells
### (mean bias calculated over the period 1950-2020 for each month)
print("remove slpressure bias")
diff = read.table("input/diff_station_grid_slp.txt",head=T,sep=";")
diff[,1] <- as.character(diff[,1])
diff = diff[which(substr(diff$abbr_var,1,3)%in%substr(stationset,1,4)),]
for(j in 2:ncol(station_data)){
  for(i in c(1:12)){
    station_data[month(station_data$date)==i,j]<-station_data[month(station_data$date)==i,j]-diff[j-1,i+1]
  }
}
#station_data <- station_data[grepl("-01-|-02-|-11-|-12-", station_data$date),] # winter months I want to reconstruct (NOV-FEB)

### subset raster data
targetdays <- paste("X",gsub("-",".",analog.dates$date),sep="")
slp_kalman <- tabs[[1:nrow(analog.dates)]]
names(slp_kalman) <- targetdays
slp_kalman_anom <- tabs_anom[[1:nrow(analog.dates)]]
names(slp_kalman_anom) <- targetdays


#############################################################################################################
### 2. KALMAN FITTING
#############################################################################################################

#----------------------------------------------------------------------------------------
### 2.1 DEFINE VALUES
#----------------------------------------------------------------------------------------
### define horizontal (L) and vertical (Z) localisation parameters (sigma in km (L) and m (Z))
if (kalmanset$localize == T)  {
  L = kalmanset$cutoff
  Z = kalmanset$Z_cutoff}
### define measurement error
R <- kalmanset$R

#----------------------------------------------------------------------------------------
### 2.2 CALCULATE DISTANCE MATRIX
#----------------------------------------------------------------------------------------
if (kalmanset$localize ==T) {
print("calculate distance matrix")

EU.alt_vals <- data.frame(rasterToPoints(EU.alt))
colnames(EU.alt_vals) <- c("x","y","layer")
mat1 <- matrix(rep(EU.alt_vals$layer, each=length(station_coords$Station)),nrow= length(station_coords$Station),ncol=length(EU.alt_vals$layer))
stat <- cellFromXY(tabs_anom[[1]], cbind(as.numeric(unlist(station_coords[,6])), as.numeric(unlist(station_coords[,5]))))
cell.nrs <- which(!is.na(values(tabs_anom[[1]])))
h.help <- match(stat,cell.nrs)
mat2 <- matrix(rep(EU.alt_vals$layer[h.help], each=length(EU.alt_vals$layer)),nrow= length(station_coords$Station),ncol=length(EU.alt_vals$layer), byrow=T)
alt.mat <- mat1-mat2
#View(alt.mat[1:41,h.help])

### Create Distance matrix for inraster
# get lat lon of stations (for slp: 41 in Nov-Dec and 34 in Jan-Feb) and their distance to the 23517 gridcells
stn_coord <- matrix(cbind(as.numeric(unlist(station_coords[,6])),as.numeric(unlist(station_coords[,5]))), ncol=2) # 1:41
msk_coord <- EU.alt_vals[,1:2] # 1:23517, coordinate points that aren't NA
Dist <- list()
for (i in 1:nrow(msk_coord)) {
  Dist[[i]] <- spDists(stn_coord, msk_coord[i,], longlat = T)
}
Dist <- Reduce(cbind, Dist)
#View(Dist[1:41,h.help])
}

#----------------------------------------------------------------------------------------
### 2.3 Kalman Filter for each day in analogue dates
#----------------------------------------------------------------------------------------
########################################################################################
#### parallelize here
########################################################################################
for (day in 1:nrow(analog.dates)) {

  print(analog.dates$date[day])
  print("subset raster & station data")
  
  ### subset raster for analog days
  analogdays <- as.Date(as.numeric(analog.dates[day,-1]),origin = "1970-01-01")
  analogdays <- paste("X",gsub("-",".",analogdays),sep="")
  slp_analogs_anom <- subset(tabs_anom,analogdays)
  
  ### define number of ensembles (evtl. change in kalman filter script)
  nens <- length(analogdays)
  
  ### create data matrix of analog days and add climatology (slp_avg)
  rast_vals <- rasterToPoints(slp_analogs_anom)
  doy <- paste("X0000.",gsub("-",".",substr(analog.dates$date[day],6,10)),sep="")
  clim <- climatology_vals[,doy]
  rast_vals[,-c(1:2)] <- apply(rast_vals[,-c(1:2)],2,function(x) x+clim)
  rast_vals_filtered <- rast_vals #for update
  
  ### subset station data and create observation vector
  station_data_sub <- station_data[station_data$date==analog.dates$date[day],]
  coords_sub <- station_coords[station_coords$abbr.%in%substr(names(station_data_sub)[!is.na(station_data_sub)][-1],1,3),]
  obs <- as.numeric(station_data_sub[,match(coords_sub$abbr., substr(colnames(station_data_sub[,-1]),1,3))+1])
  
  ### create H vector
  stat <- cellFromXY(slp_analogs_anom[[1]], cbind(as.numeric(unlist(coords_sub[,6])), as.numeric(unlist(coords_sub[,5]))))
  cell.nrs<-which(!is.na(values(slp_analogs_anom[[1]])))
  h.i <- match(stat,cell.nrs)
  H <- t(sapply(h.i, function(i) {out <- rep(0, length(cell.nrs)) ; out[i] <- 1; out}))
  
  ### create observation covariance matrix
  Rcal <- rep(R,length(stat))

  ### define background mean and variance
  xb <- rast_vals[,-c(1:2)]
  xb_mean <- rowMeans(xb)
  xb_1 <- xb - xb_mean
  
  ### calculate covariance matrix
  print("calculate covariance matrix")
  
  if(kalmanset$localize==T) {
    PH <- matrix(ncol=length(obs),nrow = length(cell.nrs))
    for (j in seq_along(obs)){
      PH[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))*exp(-(((abs(alt.mat[j,]))^2)/(2*Z^2)))*exp(-(((abs(Dist[j,]))^2)/(2*L^2)))
    }
  } else {
    PH <- matrix(ncol=length(obs),nrow = length(cell.nrs))
    for (j in seq_along(obs)){
      PH[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))
    }
  }
  
  ### do Kalman fitting
  print("apply Kalman fitting")
  HPHR <- (H %*% PH + diag(Rcal))
  K <- PH %*% solve(HPHR)
  eHPHR <- eigen(HPHR)
  sqrtHPHR <- eHPHR$vectors %*% diag(sqrt(eHPHR$values)) %*% t(eHPHR$vectors)
  Ktilde <- PH %*% t(solve(sqrtHPHR)) %*% solve(sqrtHPHR + diag(sqrt(Rcal)))
  xa <- as.vector(xb_mean + K %*% (obs - (H %*% xb_mean)))
  
  xb_1_small <- xb_1[h.i,]
  a_small <- Ktilde %*% H[,h.i]         #[erstes Matrixprodukt]
  b <- a_small %*% xb_1_small       #[zweites Matrixprodukt]
  xa1 <- xb_1 - b
  analysis <- xa + xa1
  # xa1 <- xb_1 - Ktilde %*% H %*% xb_1
  # analysis <- xa + xa1
  
  ### write data to raster files (pressure & pressure anom)
  analysis_anom <- apply(analysis,2,function(x) x-clim)
  values(slp_kalman[[day]])[!is.na(values(slp_kalman[[day]]))]<- as.numeric(analysis[,1])
  values(slp_kalman_anom[[day]])[!is.na(values(slp_kalman_anom[[day]]))]<- as.numeric(analysis_anom[,1])
}


#############################################################################################################
### 3. WRITE RESULTS TO FILES
#############################################################################################################
sdate <- as.character(analog.dates[1,1])
edate <- as.character(analog.dates[nrow(analog.dates),1])

if (kalmanset$localize==T) {
  props1 <- "_localized"
  props2 <- paste("_",kalmanset$cutoff,"km",sep="")
  props3 <- paste("_",kalmanset$Z_cutoff,"m",sep="")
} else {
  props1 <- c()
  props2 <- c()
  props3 <- c()
}

# set Z dimension to time
slp_kalman <- setZ(slp_kalman,seq(as.Date(sdate),as.Date(edate),by="day")
                           [grepl("-01-|-02-|-11-|-12-", seq(as.Date(sdate),as.Date(edate),by="day"))],"time")
slp_kalman_anom <- setZ(slp_kalman_anom,seq(as.Date(sdate),as.Date(edate),by="day")
                           [grepl("-01-|-02-|-11-|-12-", seq(as.Date(sdate),as.Date(edate),by="day"))],"time")

# write gridded files
print("write raster: slp_EnKF")
writeRaster(slp_kalman,file=paste(outfolder,Sys.Date(),"_slp_EnKF_",sdate,"-",edate,props1,props2,props3,".nc",sep=""),
            overwrite=T, varname="slp", varunit="hPa",longname="daily mean sea level pressure",zname="time",xname="lon",yname="lat",)
print("write raster: slp_EnKF anomalies")
writeRaster(slp_kalman_anom,file=paste(outfolder,Sys.Date(),"_slp_EnKF_anom_",sdate,"-",edate,props1,props2,props3,".nc",sep=""),
      overwrite=T, varname="slp", varunit="hPa",longname="daily mean sea level pressure anomalies",zname="time",xname="lon",yname="lat")

readme<-data.frame("description"=c("kalman filter","period","analog dates","station data","station coordinates","input raster: raw","input raster: deseas","input raster: climatology","station set","localize","L [if localize ==T]","Z [if localize ==T]","R"),"parameters"=NA)
readme[1,2]<-as.character(as.Date(Sys.Date()))
readme[2,2]<-paste(sdate,edate,sep="-")
for(i in 3:nrow(readme)) {readme[i,2]<-as.character((paste(kalmanset[[i-2]], collapse = ", ")))}
readme[9,2]<-paste(stationset,collapse = ", ")
write.table(readme,file=paste(outfolder,Sys.Date(),"_readme_EnKF_",sdate,"-",edate,props1,props2,props3,".txt",sep=""))

mtm <- proc.time()
print(paste("calculation time EnKF:",(mtm[3]-ptm[3])%/%60,"minutes", round((mtm[3]-ptm[3])%%60),"seconds" ,sep=" "))
ptm <- mtm

# remove variables from environment
if(kalmanset$localize==T){rm(EU.alt_vals,alt.mat,Dist,L,Z)}
rm(sdate,edate,readme,analysis,analysis_anom,climatology_vals,coords_sub,diff,eHPHR,H,HPHR,K,kalmanset,
   Ktilde,PH,rast_vals,rast_vals_filtered,sqrtHPHR,station_coords,station_data,station_data_sub,tabs,
   tabs_climatology,slp_analogs_anom,slp_kalman,slp_kalman_anom,TOT,xa1,xb,xb_1,analogdays,cell.nrs,
   clim,day,doy,h.i,i,in_year,j,mtm,nens,obs,outfolder,props1,props2,props3,ptm,R,Rcal,stat,stationset,
   targetdays,xa,xb_mean,EU.alt,tabs_anom,analog.dates)

# END

