### 05 EnKF
### Ensemble Kalman fitting (temperature)
### 17.08.2021

### input needed:
# analogue dates (with n best analogues); should be an ensemble of min 30 Analogues
# 3 temperature raster datasets (eobs, eobs_deseas, eobs_climatology)
# station measurements (TOT)
# station coordinates ("Data/Station_data/coordinates.txt")
# monthly difference between station measurements and corresponding gridcell of EOBS
# altitude data for vertical localisation
### options:
# vertical and horizontal localisation (+ cutoff distances), station selection, set observation error R
### packages:
# raster, gdata

#############################################################################################################
# from console (per year): Rscript 05_EnKF.R 1788
# for many years, use bash script for parallel runs (./EnKF_parallel.sh)
#############################################################################################################

#############################################################################################################
#############################################################################################################
# 1. READ DATA
#    1.1 READ ANALOGUE DATES
#    1.2 READ RASTER DATA
#    1.3 READ STATION DATA
# 2. KALMAN FITTING
#    2.1 DEFINE VALUES
#    2.2 CALCULATE DISTANCE MATRIX
#    2.3 Kalman Filter for each day in analogue dates
# 3. WRITE RESULTS TO FILES
#############################################################################################################
#############################################################################################################

# get time
ptm <- proc.time()
#in_year=1789
### the following lines take arguments (year) from console
args <- commandArgs(TRUE)
if (length(args)>0) {
  in_year = as.numeric(args[1])
  
  # load other scripts
  source("D01_packages.R")
  
  # define kalmanset
  kalmanset <- list(
    analog_dates = "output/ARM_run_2021-09-06/2021-09-06_analog.dates_1788-11-01-2020-12-31_deseas_detrend_RMSE.txt", # define analog dates
    station_data = "input/TOT_homhist.txt",      # define station data (TOT)
    station_xy = "input/coordinates.xlsx", # define station coordinate file
    input_raster = "grid/eobs_detrend_offs_masked.nc", # define input raster (eobs, rest calculated with cdo commands)
    input_raster_deseas = "grid/eobs_deseas_detrend_offs.nc",
    climatology = "grid/eobs_ydaymean.nc",
    # define set of stations to do kalman filter
    stationset = c("Nov-Dec"), # define set of stations to do kalman filter
    #stationset = c("Jan-Feb"),
    localize = T,
    cutoff = 750, # define horizontal cutoff distance L [km] (if localize = T)
    Z_cutoff = 1000, # define vertical cutoff distance z [m]  (if localize = T)
    R = 2.2 # define measurement error R [K]
  )
}

if (!exists("in_year")) message("define year for which EnKF should be run")
stopifnot(exists("in_year"))

print(in_year)

## create output folder for validation
subDir <- paste("EnKF",Sys.Date(),sep="_temp_")
dir.create(file.path(dirname(kalmanset$analog_dates),subDir), showWarnings = F)
outfolder <- paste(file.path(dirname(kalmanset$analog_dates),subDir),"/",sep="")
rm(subDir)

#############################################################################################################
### 1. READ DATA
#############################################################################################################

#----------------------------------------------------------------------------------------
### 1.1 READ ANALOGUE DATES
#----------------------------------------------------------------------------------------
### load analogue dates
analog.dates <- read.table(kalmanset$analog_dates,as.is=T)

#----------------------------------------------------------------------------------------
### 1.2 READ RASTER DATA
#----------------------------------------------------------------------------------------
### load input rasters
suppressWarnings(tabs <- brick(kalmanset$input_raster))
suppressWarnings(tabs_deseas <- brick(kalmanset$input_raster_deseas))
suppressWarnings(tabs_climatology <- brick(kalmanset$climatology))

climatology_vals <- rasterToPoints(tabs_climatology)
substr(dimnames(climatology_vals)[[2]][3:368],2,5) <- "0000"

EU.alt <- raster("grid/alt_eobs.nc")

#----------------------------------------------------------------------------------------
### 1.3 READ STATION DATA
#----------------------------------------------------------------------------------------
if (kalmanset$stationset=="Nov-Dec") { 
   stationset <- c("BAR","BAS","BRL","BER","BOL","BSW","BUD","CBT","CET","CLF","COP","EDI","ERF","GEN","GOT",#"GDC",
                   "HOH","INN","ROC","LIL","LIV","LDN","LNB","MAN","MAR","MIL","MNT","MOS","NAN","PAD",
                   "PAR","PRK","REG","RXV","ROM","RST","ROU","STB","SPT","PET","STO","STR","TRH","TRK","UPS","VIE","VIL","WAR",
                   "ZAG","ZWA")
} else if (kalmanset$stationset=="Jan-Feb") {
  stationset <- c("BAR","BAS","BER","BOL","BSW","BUD","CBT","CET","CLF","EDI","GEN","GOT",#"GDC",
                  "HOH","INN","KEZ","ROC","LIL","LIV","LDN","MAN","MAR","MIL","MNT","MOS","NAN","PAD",
                  "PAR","PRK","REG","RXV","ROM","RST","ROU","STB","PET","STO","TRH","TRK","UPS","VIE","VIL","WAR",
                  "ZAG","ZWA")  
} else {
  stationset <- kalmanset$stationset
}


### Read station pressure measurements
TOT <- read.table(kalmanset$station_data,header=T,as.is=T)
station_data <- data.frame("date"=as.Date(TOT$date),TOT[,names(TOT)%in%as.vector(matchcols(TOT, with="_ta", method = "or"))])
station_data <- data.frame("date"=station_data$date,station_data[,names(station_data)%in%as.vector(matchcols(station_data, with=stationset, method = "or"))])
stationset <- substr(colnames(station_data[,-1]),1,3)

#------------------------------------------------------------------------------------------------
## for console run
if (exists("in_year")) {
  if (kalmanset$stationset=="Jan-Feb") {
    station_data <- station_data[which(year(station_data$date)%in%in_year & grepl("-01-|-02-",station_data$date)),]
    analog.dates <- analog.dates[which(substr(analog.dates$date,1,4)%in%in_year & grepl("-01-|-02-",analog.dates$date)),]
  } else if (kalmanset$stationset=="Nov-Dec") {
    station_data <- station_data[which(year(station_data$date)%in%in_year & grepl("-11-|-12-",station_data$date)),]
    analog.dates <- analog.dates[which(substr(analog.dates$date,1,4)%in%in_year & grepl("-11-|-12-",analog.dates$date)),]
  }
}
#------------------------------------------------------------------------------------------------

### Read station coordinates
station_coords <- read_excel(kalmanset$station_xy,col_names=TRUE,skip=0) # make sure readxl package is loaded
station_coords <- as.data.frame(station_coords[which(station_coords$abbr.%in%stationset),])
station_coords <- station_coords[with(station_coords, order(abbr.)),]
if (kalmanset$stationset=="Jan-Feb") {
  station_coords[grepl("Gordon-Castle",station_coords$Station),"Lat"]<-station_coords[grepl("Gordon-Castle",station_coords$Station),"Lat"]-0.2
  station_coords[grepl("Trondheim",station_coords$Station),"Lat"]<-station_coords[grepl("Trondheim",station_coords$Station),"Lat"]-0.15
} else if (kalmanset$stationset=="Nov-Dec") {
  station_coords[grepl("Gordon-Castle",station_coords$Station),"Lat"]<-station_coords[grepl("Gordon-Castle",station_coords$Station),"Lat"]-0.2
  station_coords[grepl("Trondheim",station_coords$Station),"Lat"]<-station_coords[grepl("Trondheim",station_coords$Station),"Lat"]-0.15
  station_coords[grepl("Middelburg",station_coords$Station),"Lon"]<-station_coords[grepl("Trondheim",station_coords$Station),"Lon"]+0.09
} # these small adjustments are made to match with an EOBS gridcell, otherwise they are NA (not on the E-OBS dataset because too close to the shore)


### remove bias between temperature measurements and gridcells
### (mean bias calculated over the period 1950-2020 for each month)
print("remove temperature bias")
diff = read.table("input/diff_station_grid_ta_TEST.txt",head=T,sep=";")
diff[,1] <- as.character(diff[,1])
diff = diff[which(substr(diff$abbr_var,1,3)%in%substr(stationset,1,4)),]
for(j in 2:ncol(station_data)){
  for(i in c(1:12)){
    station_data[month(station_data$date)==i,j]<-station_data[month(station_data$date)==i,j]-diff[j-1,i+1]
  }
}

### subset raster data
targetdays <- paste("X",gsub("-",".",analog.dates$date),sep="")
temp_kalman <- tabs[[1:nrow(analog.dates)]]
names(temp_kalman) <- targetdays
temp_kalman_deseas <- temp_kalman


#############################################################################################################
### 2. KALMAN FITTING
#############################################################################################################

#----------------------------------------------------------------------------------------
### 2.1 DEFINE VALUES
#----------------------------------------------------------------------------------------
### define horizontal (L) and vertical (Z) localisation parameters (sigma in km (L) and m (Z))
if (kalmanset$localize == T)  {
  L = kalmanset$cutoff
  Z = kalmanset$Z_cutoff}
### define measurement error
R <- kalmanset$R

#----------------------------------------------------------------------------------------
### 2.2 CALCULATE DISTANCE MATRIX
#----------------------------------------------------------------------------------------
if (kalmanset$localize ==T) {
print("calculate distance matrix")

EU.alt_vals <- data.frame(rasterToPoints(EU.alt))
colnames(EU.alt_vals) <- c("x","y","layer")
mat1 <- matrix(rep(EU.alt_vals$layer, each=length(station_coords$Station)),nrow= length(station_coords$Station),ncol=length(EU.alt_vals$layer))
stat <- cellFromXY(tabs_deseas[[1]], cbind(as.numeric(unlist(station_coords[,6])), as.numeric(unlist(station_coords[,5]))))
cell.nrs <- which(!is.na(values(tabs_deseas[[1]])))
h.help <- match(stat,cell.nrs)
mat2 <- matrix(rep(EU.alt_vals$layer[h.help], each=length(EU.alt_vals$layer)),nrow= length(station_coords$Station),ncol=length(EU.alt_vals$layer), byrow=T)
alt.mat <- mat1-mat2
#View(alt.mat[1:51,h.help]) 

### Create Distance matrix for inraster
# get lat lon of stations (for temp: 51 in Nov-Dec and 44 in Jan-Feb)
stn_coord <- matrix(cbind(as.numeric(unlist(station_coords[,6])),as.numeric(unlist(station_coords[,5]))), ncol=2) # 1:43
msk_coord <- EU.alt_vals[,1:2]
Dist <- list()
for (i in 1:nrow(msk_coord)) {
  Dist[[i]] <- spDists(stn_coord, msk_coord[i,], longlat = T)
}
Dist <- Reduce(cbind, Dist)
#View(Dist[1:51,h.help])
}

#----------------------------------------------------------------------------------------
### 2.3 Kalman Filter for each day in analogue dates
#----------------------------------------------------------------------------------------
########################################################################################
#### parallelize here
########################################################################################
for (day in 1:nrow(analog.dates)) {

  print(analog.dates$date[day])
  print("subset raster & station data")
  
  ### subset raster for analog days
  analogdays <- as.Date(as.numeric(analog.dates[day,-1]),origin = "1970-01-01")
  analogdays <- paste("X",gsub("-",".",analogdays),sep="")
  temp_analogs_deseas <- subset(tabs_deseas,analogdays)
  
  ### define number of ensembles (evtl. change in kalman filter script)
  nens <- length(analogdays)
  
  ### create data matrix of analog days and add climatology
  rast_vals <- rasterToPoints(temp_analogs_deseas)
  doy <- paste("X0000.",gsub("-",".",substr(analog.dates$date[day],6,10)),sep="")
  clim <- climatology_vals[,doy]
  rast_vals[,-c(1:2)] <- apply(rast_vals[,-c(1:2)],2,function(x) x+clim)
  rast_vals_filtered <- rast_vals #for update
  
  ### subset station data and create observation vector
  station_data_sub <- station_data[station_data$date==analog.dates$date[day],]
  coords_sub <- station_coords[station_coords$abbr.%in%substr(names(station_data_sub)[!is.na(station_data_sub)][-1],1,3),]
  obs <- as.numeric(station_data_sub[,match(coords_sub$abbr., substr(colnames(station_data_sub[,-1]),1,3))+1])
  
  ### create H vector
  stat <- cellFromXY(temp_analogs_deseas[[1]], cbind(as.numeric(unlist(coords_sub[,6])), as.numeric(unlist(coords_sub[,5]))))
  cell.nrs<-which(!is.na(values(temp_analogs_deseas[[1]])))
  h.i <- match(stat,cell.nrs)
  H <- t(sapply(h.i, function(i) {out <- rep(0, length(cell.nrs)) ; out[i] <- 1; out}))
  
  ### create observation covariance matrix
  Rcal <- rep(R,length(stat))

  ### define background mean and variance
  xb <- rast_vals[,-c(1:2)]
  xb_mean <- rowMeans(xb)
  xb_1 <- xb - xb_mean
  
  ### calculate covariance matrix
  print("calculate covariance matrix")
  
  if(kalmanset$localize==T) {
    PH <- matrix(ncol=length(obs),nrow = length(cell.nrs))
    for (j in seq_along(obs)){
      PH[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))*exp(-(((abs(alt.mat[j,]))^2)/(2*Z^2)))*exp(-(((abs(Dist[j,]))^2)/(2*L^2)))
    }
  } else {
    PH <- matrix(ncol=length(obs),nrow = length(cell.nrs))
    for (j in seq_along(obs)){
      PH[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))
    }
  }
  
  ### do Kalman fitting
  print("apply Kalman fitting")
  HPHR <- (H %*% PH + diag(Rcal))
  K <- PH %*% solve(HPHR)
  eHPHR <- eigen(HPHR)
  sqrtHPHR <- eHPHR$vectors %*% diag(sqrt(eHPHR$values)) %*% t(eHPHR$vectors)
  Ktilde <- PH %*% t(solve(sqrtHPHR)) %*% solve(sqrtHPHR + diag(sqrt(Rcal)))
  xa <- as.vector(xb_mean + K %*% (obs - (H %*% xb_mean)))
  
  xb_1_small <- xb_1[h.i,]
  a_small <- Ktilde %*% H[,h.i]         #[erstes Matrixprodukt]
  b <- a_small %*% xb_1_small       #[zweites Matrixprodukt]
  xa1 <- xb_1 - b
  analysis <- xa + xa1
  # xa1 <- xb_1 - Ktilde %*% H %*% xb_1
  # analysis <- xa + xa1
  
  ### write data to raster files (temperature & deseas temperature)
  analysis_deseas <- apply(analysis,2,function(x) x-clim)
  values(temp_kalman[[day]])[!is.na(values(temp_kalman[[day]]))]<- as.numeric(analysis[,1])
  values(temp_kalman_deseas[[day]])[!is.na(values(temp_kalman_deseas[[day]]))]<- as.numeric(analysis_deseas[,1])
}


#############################################################################################################
### 3. WRITE RESULTS TO FILES
#############################################################################################################
sdate <- as.character(analog.dates[1,1])
edate <- as.character(analog.dates[nrow(analog.dates),1])

if (kalmanset$localize==T) {
  props1 <- "_localized"
  props2 <- paste("_",kalmanset$cutoff,"km",sep="")
  props3 <- paste("_",kalmanset$Z_cutoff,"m",sep="")
} else {
  props1 <- c()
  props2 <- c()
  props3 <- c()
}

# set Z dimension to time
temp_kalman <- setZ(temp_kalman,seq(as.Date(sdate),as.Date(edate),by="day")
                           [grepl("-01-|-02-|-11-|-12-", seq(as.Date(sdate),as.Date(edate),by="day"))],"time")
temp_kalman_deseas <- setZ(temp_kalman_deseas,seq(as.Date(sdate),as.Date(edate),by="day")
                           [grepl("-01-|-02-|-11-|-12-", seq(as.Date(sdate),as.Date(edate),by="day"))],"time")

# write gridded files
print("write raster: temp_EnKF")
writeRaster(temp_kalman,file=paste(outfolder,Sys.Date(),"_temp_EnKF_",sdate,"-",edate,props1,props2,props3,".nc",sep=""),
            overwrite=T, varname="temp", varunit="degC",longname="daily mean temperature",zname="time",xname="lon",yname="lat")
print("write raster: temp_EnKF deseasonalised")
writeRaster(temp_kalman_deseas,file=paste(outfolder,Sys.Date(),"_temp_EnKF_deseas_",sdate,"-",edate,props1,props2,props3,".nc",sep=""),
      overwrite=T, varname="temp", varunit="degC",longname="daily mean temperature anomalies",zname="time",xname="lon",yname="lat")

readme<-data.frame("description"=c("kalman filter","period","analog dates","station data","station coordinates","input raster: raw","input raster: deseas","input raster: climatology","station set","localize","L [if localize ==T]","Z [if localize ==T]","R"),"parameters"=NA)
readme[1,2]<-as.character(as.Date(Sys.Date()))
readme[2,2]<-paste(sdate,edate,sep="-")
for(i in 3:nrow(readme)) {readme[i,2]<-as.character((paste(kalmanset[[i-2]], collapse = ", ")))}
readme[9,2]<-paste(stationset,collapse = ", ")
write.table(readme,file=paste(outfolder,Sys.Date(),"_readme_EnKF_",sdate,"-",edate,props1,props2,props3,".txt",sep=""))

mtm <- proc.time()
print(paste("calculation time EnKF:",(mtm[3]-ptm[3])%/%60,"minutes", round((mtm[3]-ptm[3])%%60),"seconds" ,sep=" "))
ptm <- mtm

# remove variables from environment
if(kalmanset$localize==T){rm(EU.alt_vals,alt.mat,Dist,L,Z)}
rm(sdate,edate,readme,analysis,analysis_deseas,climatology_vals,coords_sub,diff,eHPHR,H,HPHR,K,kalmanset,
   Ktilde,PH,rast_vals,rast_vals_filtered,sqrtHPHR,station_coords,station_data,station_data_sub,tabs,
   tabs_climatology,temp_analogs_deseas,temp_kalman,temp_kalman_deseas,TOT,xa1,xb,xb_1,analogdays,cell.nrs,
   clim,day,doy,h.i,i,in_year,j,mtm,nens,obs,outfolder,props1,props2,props3,ptm,R,Rcal,stat,stationset,
   targetdays,xa,xb_mean,EU.alt,tabs_deseas,analog.dates)

# END

